"""
Load fuel prices needed for the model
"""

from asyncio.log import logger
from typing import Dict, List

import pandas as pd

from powergenome.eia_opendata import add_user_fuel_prices


def fuel_cost_table(
    fuel_costs: pd.DataFrame,
    generators: pd.DataFrame,
    settings: dict,
    num_hours: int = None,
) -> pd.DataFrame:
    """Create a table of fuel costs formatted for the GenX model.

    Costs are based on the `fuel_costs` dataframe and any values listed in the settings
    dictionary under the key "user_fuel_price". If CCS fuels or a carbon tax are defined
    in the settings then the

    Parameters
    ----------
    fuel_costs : pd.DataFrame
        A table of fuel prices. Must have columns "year", "price", "fuel", "region", and
        "full_fuel_name". "fuel" should be a base fuel name such as coal or distillate and
        cannot include an underscore. If the full fuel name is a combination of the region,
        a scenario, and the base fuel name, the base fuel name should be the last element
        so that it is selected when the string is split on underscores.

        >>> <full_fuel_name>.split("_")[-1] = <fuel>
    generators : pd.DataFrame
        A table of generators with the column "Fuel". The values in this column should
        correspond to either the "full_fuel_name" column or one of the fuels in the
        settings key "user_fuel_price". If regional prices are provided in the settings
        then the fuel name should be <region>_<fuel>.
    settings : dict
        Should include the key "fuel_emission_factors" with CO2 emissions in tonnes
        per MMBTU for each fuel type used.

        If adding user prices, should have the key "user_fuel_price" with value of a
        dictionary matching user fuel names and prices. Prices can either be a single
        price for all regions or a price per region. For example this shows biomass with
        different prices in two regions and ZCF with the same price in all regions:

        settings["user_fuel_price"] = {
            "biomass": {"SC_VACA": 10, "PJM_DOM": 5},
            "ZCF": 15
        }

        If the keys "target_usd_year" and "user_fuel_usd_year" are also included, fuel
        prices will be corrected to the correct USD year. "user_fuel_usd_year" should
        be a dictionary with fuel name: USD year pairings. Only fuels included in this
        dictionary will have their prices changed to the target USD year.

    Returns
    -------
    pd.DataFrame
        The cost of fuels used by generators in the final year of a modeling period.
        Formatted for GenX, where headers are the fuel names, the first row is the
        fuel CO2 content (tonnes per MMBTU), and subsequent rows are hourly prices.
        Prices are identical in all hours. The first (index) column has the header
        "Time_Index" and values from 0-N, where N is the number of hours used in the model.
    """
    all_fuel_costs = add_user_fuel_prices(settings, fuel_costs)
    unique_fuels = generators["Fuel"].drop_duplicates()
    model_year_costs = all_fuel_costs.loc[
        all_fuel_costs["year"] == settings["model_year"], :
    ]
    fuel_df = pd.DataFrame(unique_fuels)

    fuel_price_map = {
        row.full_fuel_name: row.price
        for row in model_year_costs.itertuples(index=False, name="row")
    }

    emission_dict = settings.get("fuel_emission_factors", {}) or {}
    user_fuels = set(all_fuel_costs["fuel"]) - set(fuel_costs["fuel"])
    for u_f in user_fuels:
        if u_f not in emission_dict.keys():
            logger.warning(
                "\n\n**********************\n"
                f"The user fuel {u_f} does not have an emissions factor specified in "
                "the settings parameter 'fuel_emission_factors'. This is fine if the "
                "emission factor should be 0, otherwise be sure to add a value.\n"
            )
    fuel_emission_map = {}
    for full_fuel_name in fuel_price_map:
        if (
            full_fuel_name.split("_")[-1]
            in (settings.get("aeo_fuel_scenarios", {}) or {}).keys()
        ):
            base_fuel_name = full_fuel_name.split("_")[-1]
        elif (
            full_fuel_name.split("_")[-1]
            in (settings.get("user_fuel_price", {}) or {}).keys()
        ):
            base_fuel_name = full_fuel_name.split("_")[-1]
        else:
            base_fuel_name = full_fuel_name
        if base_fuel_name in emission_dict:
            fuel_emission_map[full_fuel_name] = emission_dict[base_fuel_name]
        else:
            fuel_emission_map[full_fuel_name] = 0

    ccs_fuels = (settings.get("ccs_fuel_map", {}) or {}).values()
    for ccs_fuel in ccs_fuels:
        fuels = generators.loc[
            generators["Fuel"].str.contains(ccs_fuel), "Fuel"
        ].unique()
        for f in fuels:
            # keep the non-ccs price
            base_name = ("_").join(f.split("_")[:-1])
            fuel_price_map[f] = fuel_price_map[base_name]
            fuel_emission_map[f] = fuel_emission_map[base_name]

    fuel_df["Cost_per_MMBtu"] = fuel_df["Fuel"].map(fuel_price_map)
    fuel_df["CO2_content_tons_per_MMBtu"] = fuel_df["Fuel"].map(fuel_emission_map)

    # Slow to loop through all of the rows this way but the df shouldn't be too long
    if settings.get("co2_pipeline_filters") and settings.get("co2_pipeline_cost_fn"):
        ccs_disposal_cost = 0
    else:
        ccs_disposal_cost = settings.get("ccs_disposal_cost", 0)
    fuel_df = fuel_df.apply(
        adjust_ccs_fuels,
        axis=1,
        ccs_fuels=(settings.get("ccs_fuel_map", {}) or {}).values(),
        ccs_capture_rate=(settings.get("ccs_capture_rate", {}) or {}),
        ccs_disposal_cost=ccs_disposal_cost,
    )
    fuel_df = add_carbon_tax(fuel_df, settings.get("carbon_tax"))
    fuel_df["Cost_per_MMBtu"] = fuel_df["Cost_per_MMBtu"]
    fuel_df["CO2_content_tons_per_MMBtu"] = fuel_df["CO2_content_tons_per_MMBtu"]
    fuel_df.fillna(0, inplace=True)

    if settings.get("reduce_time_domain"):
        days = settings["time_domain_days_per_period"]
        time_periods = settings["time_domain_periods"]
        num_hours = days * time_periods * 24
    elif num_hours is None:
        num_hours = 8760

    fuel_df_prices = pd.DataFrame(
        [fuel_df["Cost_per_MMBtu"]], index=range(1, num_hours + 1)
    )
    fuel_df_prices = fuel_df_prices.round(2)
    fuel_df_prices.columns = unique_fuels

    fuel_df_top = pd.DataFrame([fuel_df["CO2_content_tons_per_MMBtu"]])
    fuel_df_top = fuel_df_top.round(5)
    fuel_df_top.columns = unique_fuels
    fuel_df_top.index = [0]

    fuel_frames = [fuel_df_top, fuel_df_prices]
    fuel_df_new = pd.concat(fuel_frames)
    fuel_df_new.index.name = "Time_Index"
    return fuel_df_new


# def modify_fuel_new_genx():


def adjust_ccs_fuels(
    ccs_fuel_row: pd.Series,
    ccs_fuels: List[str] = None,
    ccs_capture_rate: Dict[str, float] = {},
    ccs_disposal_cost: float = None,
) -> pd.Series:
    """Adjust the "CO2_content_tons_per_MMBtu" and "Cost_per_MMBtu" values to account for
    the value from settings parameter "ccs_capture_rate".

    If using this function to adjust the CO2 content and cost for CCS-specific fuels,
    the settings dict should map the names of technologies to a base CCS fuel name in the
    parameter "ccs_fuel_map". The base CCS fuel names do not include a region or scenario,
    they are something like "naturalgas_ccs90".


    Parameters
    ----------
    ccs_fuel_row : pd.Series
        A single row from the larger fuel dataframe with columns "Fuel", "Cost_per_MMBtu",
        and "CO2_content_tons_per_MMBtu".
    ccs_fuels : List[str], optional
        A list of CCS fuels mapped to generator types, by default None
    ccs_capture_rate : Dict[str, float], optional
        The capture rate (0-1) for each CCS fuel type in `ccs_fuels`, by default {}
    ccs_disposal_cost : float, optional
        The cost in USD per tonne of CO2 disposal that should be added to a fuel price,
        by default None

    Returns
    -------
    pd.Series
        If the fuel is mapped to a CCS technology, the "CO2_content_tons_per_MMBtu" and
        "Cost_per_MMBtu" values will be modified.

    Raises
    ------
    KeyError
        One of the CCS fuels mapped to a technology is not included in the "ccs_capture_rate"
        dict.
    """

    base_fuel_name = None
    for ccs_fuel in ccs_fuels or []:
        if ccs_fuel not in ccs_capture_rate.keys():
            raise KeyError(
                f"The CCS fuel name {ccs_fuel} from settings parameter 'ccs_fuel_map' "
                "does not have capture rate in the settings parameter 'ccs_capture_rate'."
                "Adjust your settings to include the capture rate or remove the fuel."
            )
        if ccs_fuel in ccs_fuel_row["Fuel"]:
            base_fuel_name = ccs_fuel
    if base_fuel_name:
        # USD/tonne disposal
        if not ccs_disposal_cost:
            logger.debug(
                "You did not specify a fuel-modifying CCS disposal cost, so it will be set to $0. "
                "Set a non-zero value with the settings parameter 'ccs_disposal_cost'."
            )
            ccs_disposal_cost = 0

        capture_rate = ccs_capture_rate.get(base_fuel_name, 0)

        co2_captured = ccs_fuel_row["CO2_content_tons_per_MMBtu"] * capture_rate

        ccs_fuel_row["CO2_content_tons_per_MMBtu"] -= co2_captured
        ccs_fuel_row["Cost_per_MMBtu"] += co2_captured * ccs_disposal_cost

    else:
        pass

    return ccs_fuel_row


def add_carbon_tax(
    fuel_df: pd.DataFrame, carbon_tax_value: float = None
) -> pd.DataFrame:
    """Increases fuel prices to account for a carbon tax

    Parameters
    ----------
    fuel_df : pd.DataFrame
        Table with columns "Cost_per_MMBtu" and "CO2_content_tons_per_MMBtu"
    carbon_tax_value : float, optional
        The carbon tax cost in USD per tonne CO2, by default None.

    Returns
    -------
    pd.DataFrame
        Modified version of input df with fuel prices increased to reflect the carbon tax.
        The df is returned unaltered if no carbon tax is provided.
    """
    if not carbon_tax_value:
        return fuel_df

    for col in ["Cost_per_MMBtu", "CO2_content_tons_per_MMBtu"]:
        if col not in fuel_df.columns:
            raise KeyError(
                f"The required column {col} is missing from your fuel dataframe. Cannot "
                "apply a carbon tax to fuel prices without this column."
            )

    fuel_df.loc[:, "Cost_per_MMBtu"] = fuel_df.loc[:, "Cost_per_MMBtu"] + (
        fuel_df.loc[:, "CO2_content_tons_per_MMBtu"] * carbon_tax_value
    )

    return fuel_df



"""
Adjust price/cost from one year to another
"""

from pathlib import Path
from typing import Union
from warnings import warn

import numpy as np
import pandas as pd

from powergenome.financials import get_cpi_data as _get_cpi_data
from powergenome.financials import (
    inflation_price_adjustment as _inflation_price_adjustment,
)
from powergenome.financials import load_cpi_data as _load_cpi_data


def get_cpi_data(start_year: int = 1980, end_year: int = None) -> pd.DataFrame:
    warn(
        "The function 'get_cpi_data' has been moved to powergenome.financials -- the "
        "location in powergenome.price_adjustement will be depreciated in a future version. "
        "Update your code to use the correct import.",
        DeprecationWarning,
        stacklevel=2,
    )

    return _get_cpi_data(start_year, end_year)


def load_cpi_data(
    reload_data: bool = False, data_path: Path = None, **kwargs
) -> pd.DataFrame:
    warn(
        "The function 'load_cpi_data' has been moved to powergenome.financials -- the "
        "location in powergenome.price_adjustement will be depreciated in a future version. "
        "Update your code to use the correct import.",
        DeprecationWarning,
        stacklevel=2,
    )

    return _load_cpi_data(reload_data, data_path, **kwargs)


def inflation_price_adjustment(
    price: Union[int, float, pd.Series, pd.DataFrame, np.ndarray],
    base_year: int,
    target_year: int,
    **kwargs,
) -> float:
    warn(
        "The function 'inflation_price_adjustment' has been moved to powergenome.financials -- the "
        "location in powergenome.price_adjustement will be depreciated in a future version. "
        "Update your code to use the correct import.",
        DeprecationWarning,
        stacklevel=2,
    )

    return _inflation_price_adjustment(price, base_year, target_year, **kwargs)



import argparse
import logging
import shutil
import sys
from datetime import datetime as dt

import pandas as pd

import powergenome
from powergenome.fuels import fuel_cost_table
from powergenome.generators import GeneratorClusters, load_ipm_shapefile
from powergenome.load_profiles import make_final_load_curves
from powergenome.params import DATA_PATHS
from powergenome.transmission import (
    agg_transmission_constraints,
    transmission_line_distance,
)
from powergenome.util import (
    get_git_hash,
    init_pudl_connection,
    load_settings,
    remove_fuel_scenario_name,
)

if not sys.warnoptions:
    import warnings

    warnings.simplefilter("ignore")


def parse_command_line(argv):
    """
    Parse command line arguments. See the -h option.

    :param argv: arguments on the command line must include caller file name.
    """
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-sf",
        "--settings_file",
        dest="settings_file",
        type=str,
        default="pudl_data_extraction.yml",
        help="Specify a YAML settings file.",
    )
    parser.add_argument(
        "-rf",
        "--results_folder",
        dest="results_folder",
        type=str,
        default=dt.now().strftime("%Y-%m-%d %H.%M.%S"),
        help="Specify the results subfolder to write output",
    )
    parser.add_argument(
        "--no-current-gens",
        dest="current_gens",
        action="store_false",
        help="Don't load and cluster current generators.",
    )
    parser.add_argument(
        # "-g",
        "--no-gens",
        dest="gens",
        # type=bool,
        action="store_false",
        # default=True,
        help="Use flag to not calculate generator clusters.",
    )
    parser.add_argument(
        # "-l",
        "--no-load",
        dest="load",
        # type=bool,
        action="store_false",
        # default=True,
        help="Calculate hourly load. If False, file will not be written.",
    )
    parser.add_argument(
        # "-t",
        "--no-transmission",
        dest="transmission",
        # type=bool,
        action="store_false",
        # default=True,
        help="Calculate transmission constraints. If False, file will not be written.",
    )
    parser.add_argument(
        "-f",
        "--no-fuel",
        dest="fuel",
        # type=bool,
        action="store_false",
        # default=True,
        help=(
            "Create fuel table. If False, file will not be written."
            " Can not be created without the generators."
        ),
    )
    parser.add_argument(
        "-s",
        "--sort-gens",
        dest="sort_gens",
        # type=bool,
        action="store_true",
        # default=True,
        help=(
            "Sort generators alphabetically within region. Existing resources will still"
            " be separate from new resources."
        ),
    )
    arguments = parser.parse_args(argv[1:])
    return arguments


def main():
    args = parse_command_line(sys.argv)

    out_folder = DATA_PATHS["results"] / args.results_folder
    DATA_PATHS["results"].mkdir(exist_ok=True)
    out_folder.mkdir(exist_ok=True)

    # Create a logger to output any messages we might have...
    logger = logging.getLogger(powergenome.__name__)
    logger.setLevel(logging.INFO)
    handler = logging.StreamHandler()
    formatter = logging.Formatter(
        # More extensive test-like formatter...
        "%(asctime)s [%(levelname)8s] %(name)s:%(lineno)s %(message)s",
        # This is the datetime format string.
        "%Y-%m-%d %H:%M:%S",
    )
    handler.setFormatter(formatter)
    logger.addHandler(handler)

    filehandler = logging.FileHandler(out_folder / "log.txt")
    filehandler.setFormatter(formatter)
    logger.addHandler(filehandler)

    git_hash = get_git_hash()
    logger.info(f"Current git hash is {git_hash}")

    logger.info("Reading settings file")
    settings = load_settings(path=args.settings_file)

    # Copy the settings file to results folder
    shutil.copy(args.settings_file, out_folder)

    logger.info("Initiating PUDL connections")
    pudl_engine, pudl_out = init_pudl_connection(freq="YS")

    # Make sure everything in model_regions is either an aggregate region
    # or an IPM region. Will need to change this once we start using non-IPM
    # regions.
    ipm_regions = pd.read_sql_table("regions_entity_epaipm", pudl_engine)[
        "region_id_epaipm"
    ]
    all_valid_regions = ipm_regions.tolist() + list(
        settings.get("region_aggregations", {})
    )
    good_regions = [region in all_valid_regions for region in settings["model_regions"]]

    # assert all(good_regions), logger.warning(
    #     "One or more model regions is not valid. Check to make sure all"
    #     "regions are either in IPM or region_aggregations in the settings YAML file."
    # )
    if not all(good_regions):
        logger.warning(
            "One or more model regions is not valid. Check to make sure all regions "
            "are either in IPM or region_aggregations in the settings YAML file."
        )

    # Sort zones in the settings to make sure they are correctly sorted everywhere.
    settings["model_regions"] = sorted(settings["model_regions"])
    zones = settings["model_regions"]
    logger.info(f"Sorted zones are {', '.join(zones)}")
    zone_num_map = {
        zone: f"{number + 1}" for zone, number in zip(zones, range(len(zones)))
    }

    if args.gens:
        gc = GeneratorClusters(
            pudl_engine=pudl_engine,
            pudl_out=pudl_out,
            settings=settings,
            current_gens=args.current_gens,
            sort_gens=args.sort_gens,
        )
        gen_clusters = gc.create_all_generators()
        gen_clusters = remove_fuel_scenario_name(gen_clusters, settings)
        gen_clusters["Zone"] = gen_clusters["region"].map(zone_num_map)

    if args.load:
        load = make_final_load_curves(pudl_engine=pudl_engine, settings=settings)
        load.columns = "Load_MW_z" + load.columns.map(zone_num_map)

    if args.transmission:
        if args.gens is False:
            model_regions_gdf = load_ipm_shapefile(settings)
        else:
            model_regions_gdf = gc.model_regions_gdf
        transmission = agg_transmission_constraints(
            pudl_engine=pudl_engine, settings=settings
        )
        transmission = transmission.pipe(
            transmission_line_distance,
            ipm_shapefile=model_regions_gdf,
            settings=settings,
            units="mile",
        )

    if args.fuel and args.gens:
        fuels = fuel_cost_table(
            fuel_costs=gc.fuel_prices, generators=gc.all_resources, settings=settings
        )
        fuels["fuel_indices"] = range(1, len(fuels) + 1)
        fuels = remove_fuel_scenario_name(fuels, settings)

    logger.info(f"Write GenX input files to {args.results_folder}")
    if args.gens:
        gen_clusters.to_csv(
            out_folder / f"generator_clusters_{args.results_folder}.csv",
            index=False,
            # float_format="%.3f",
        )
        # if args.all_units is True:
        # gc.all_units.to_csv(out_folder / f"all_units_{args.results_folder}.csv")

    if args.load:
        load.astype(int).to_csv(out_folder / f"load_curves_{args.results_folder}.csv")

    if args.transmission:
        transmission.to_csv(
            out_folder / f"transmission_constraints_{args.results_folder}.csv",
            float_format="%.1f",
        )

    if args.fuel and args.gens:
        fuels.to_csv(out_folder / f"Fuels_data_{args.results_folder}.csv", index=False)


if __name__ == "__main__":
    main()



"Functions to cluster or otherwise reduce the number of hours in generation and load profiles"

import datetime
import logging
from typing import List, Tuple

import numpy as np
import pandas as pd
from sklearn.cluster import KMeans
from sklearn.preprocessing import minmax_scale

logger = logging.getLogger(__name__)


def max_rep_periods(
    resource_profiles: pd.DataFrame,
    load_profiles: pd.DataFrame,
    days_in_group: int,
    num_clusters: int,
) -> Tuple[pd.DataFrame, pd.DataFrame, List[int], pd.DataFrame, pd.DataFrame]:
    """Shortcut clustering when every representative period is assigned once (e.g. 52 rep
    weeks in a year).

    Parameters
    ----------
    resource_profiles : pd.DataFrame
        Hourly profiles of all resources
    load_profiles : pd.DataFrame
        Hourly demand profile in each region
    days_in_group : int
        Number of days in each period
    num_clusters : int
        Numer of representative periods

    Returns
    -------
    Tuple[pd.DataFrame, pd.DataFrame, List[int], pd.DataFrame, pd.DataFrame]
        Load and resource profile dataframes with first N hours, where N is
        num_clusters * days_in_group * 24, the cluster weights (all values are 1),
        a mapping of each period to the representative period and the month,
        and the order of representative periods.
    """
    num_hours = num_clusters * days_in_group * 24
    cluster_weights = [1] * num_clusters
    time_series_mapping = pd.DataFrame(
        data={
            "Period_Index": range(1, 1 + num_clusters),
            "Rep_Period_Index": range(1, 1 + num_clusters),
            "Month": 0,
        }
    )
    for Period_Index in time_series_mapping["Period_Index"]:
        dayOfYear = days_in_group * Period_Index
        d = datetime.datetime.strptime("{} {}".format(dayOfYear, 2011), "%j %Y")
        time_series_mapping["Month"][Period_Index - 1] = d.month
    rep_point = [f"p{i}" for i in range(1, 1 + num_clusters)]
    rep_point_df = pd.DataFrame(data=rep_point, columns=["slot"])

    reduced_load_profiles = load_profiles.iloc[:num_hours, :].reset_index(drop=True)
    reduced_resouce_profiles = resource_profiles.iloc[:num_hours, :]

    return (
        reduced_load_profiles,
        reduced_resouce_profiles,
        cluster_weights,
        time_series_mapping,
        rep_point_df,
    )


def kmeans_time_clustering(
    resource_profiles,
    load_profiles,
    days_in_group,
    num_clusters,
    include_peak_day=True,
    load_weight=1,
    variable_resources_only=True,
    n_init=100,
):
    """Reduce the number of hours in load and resource variability timeseries using
    kmeans clustering.

    This script is adapted from work originally created by Dharik Mallapragada. For more
    information see:
    - Mallapragada, D. S., Papageorgiou, D. J., Venkatesh, A., Lara, C. L., & Grossmann,
    I. E. (2018). Impact of model resolution on scenario outcomes for electricity sector
    system expansion. Energy, 163, 1231–1244.
    https://doi.org/10.1016/j.energy.2018.08.015
    - Mallapragada, D. S., Sepulveda, N. A., & Jenkins, J. D. (2020). Long-run system
    value of battery energy storage in future grids with increasing wind and solar
    generation. Applied Energy, 275, 115390.
    https://doi.org/10.1016/j.apenergy.2020.115390


    Parameters
    ----------
    resource_profiles : DataFrame
        Hourly generation profiles for all resources. Each column is a resource with
        a unique name, each row is a consecutive hour.
    load_profiles : DataFrame
        Hourly demand profiles of load. Each column is a region with a unique name. each
        row is a consecutive hour.
    days_in_group : int
        The number of 24 hour periods included in each group/cluster
    num_clusters : int
        The number of clusters to include in the output
    include_peak_day : bool, optional
        If the days with system peak demand should be included in outputs, by default
        True
    load_weight : int, optional
        A weighting factor for load profiles during clustering, by default 1
    variable_resources_only : bool, optional
        If clustering should only consider resources with variable (non-zero standard
        deviation) profiles, by default True
    n_init : int, optional
        Parameter for k-means clustering.

    Returns
    -------
    (dict, list, list)
        This function returns multiple items. The dict has keys ['load_profiles',
        'resource_profiles', 'ClusterWeights', 'AnnualGenScaleFactor', 'RMSE', and
        'AnnualProfile']

        The first list has strings with the order of periods selected e.g. ['p42','p26',
        'p3', 'p13', 'p32', 'p8'].

        The second list has integer weights of each cluster.
    """
    logger.info("Reducing time domain from 8760 hours to representative periods")
    # In cases where each cluster is selected exactly once, skip the clustering entirely
    total_cluster_hours = days_in_group * num_clusters * 24
    if len(load_profiles) - total_cluster_hours < days_in_group * 24:
        (
            load_df,
            resource_df,
            EachClusterWeight,
            time_series_mapping,
            EachClusterRepPoint,
        ) = max_rep_periods(
            resource_profiles=resource_profiles,
            load_profiles=load_profiles,
            days_in_group=days_in_group,
            num_clusters=num_clusters,
        )
        rep_period_map = {
            i + 1: int(p[1:]) for i, p in enumerate(EachClusterRepPoint["slot"])
        }
        time_series_mapping["Rep_Period"] = time_series_mapping["Rep_Period_Index"].map(
            rep_period_map
        )
        return (
            {
                "load_profiles": load_df,  # Scaled Output Load and Renewables profiles for the sampled representative groupings
                "resource_profiles": resource_df,
                "ClusterWeights": EachClusterWeight,  # Weight of each for the representative groupings
                "AnnualGenScaleFactor": 1,  # Scale factor used to adjust load output to match annual generation of original data
                "RMSE": None,  # Root mean square error between full year data and modeled full year data (duration curves)
                "AnnualProfile": None,
                "time_series_mapping": time_series_mapping,
            },
            EachClusterRepPoint,
            EachClusterWeight,
        )

    resource_col_names = resource_profiles.columns
    if variable_resources_only:
        input_std = resource_profiles.describe().loc["std", :]
        var_col_names = input_std[input_std > 0].index.to_list()
        resource_profiles = resource_profiles.loc[:, var_col_names]

    # Initialize dataframes to store final and intermediate data in

    input_data = pd.concat(
        [
            load_profiles.reset_index(drop=True),
            resource_profiles.reset_index(drop=True),
        ],
        axis=1,
    )
    input_data = input_data.reset_index(drop=True)
    original_col_names = input_data.columns.tolist()
    # CAUTION: Load Column lables should be named with the phrase "Load_"
    load_col_names = load_profiles.columns

    # Columns to be reported in output files
    new_col_names = input_data.columns.tolist() + ["GrpWeight"]
    # Dataframe storing final outputs
    final_output_data = pd.DataFrame(columns=new_col_names)

    # Dataframe storing normalized inputs
    norm_tseries = pd.DataFrame(columns=original_col_names)

    hours_per_year = len(input_data)

    # Normalized all load and renewables data 0 and LoadWeight, All Renewables b/w 0
    # and 1
    norm_tseries = pd.DataFrame(
        data=minmax_scale(input_data), columns=input_data.columns
    )
    norm_tseries.loc[:, load_col_names] *= load_weight

    # Identify hour with maximum system wide load
    hr_maxSysLoad = (
        input_data.loc[: total_cluster_hours - 1, load_col_names].sum(axis=1).idxmax()
    )
    ################################ pre-processing data to create concatenated column
    # of load, pv and wind data

    # Number of such samples in a year - by avoiding division by float we are excluding
    # a few days in each sample set
    # Hence annual generation for each zone will not exactly match up with raw data
    # num_data_points = round(hours_per_year / 24 / days_in_group)
    num_data_points = int(hours_per_year / 24 // days_in_group)

    DataPointAsColumns = [f"p{j}" for j in range(1, num_data_points + 1)]

    # Create a dictionary storing groups of time periods to average over for each hour
    HourlyGroupings = {
        i: [j for j in range(days_in_group * 24 * (i - 1), days_in_group * 24 * i)]
        for i in range(1, num_data_points + 1)
    }

    #  Create a new dataframe storing aggregated load and renewables time series
    ModifiedDataNormalized = pd.DataFrame(columns=DataPointAsColumns)
    # Original data organized in concatenated column
    ModifiedData = pd.DataFrame(columns=DataPointAsColumns)

    # Creating the dataframe with concatenated columns
    for j in range(num_data_points):
        if j == 1:  # Store  variable names for the concatenated column
            ConcatenatedRowNames = norm_tseries.loc[HourlyGroupings[j + 1], :].melt(
                id_vars=None
            )["variable"]

        ModifiedDataNormalized[DataPointAsColumns[j]] = norm_tseries.loc[
            HourlyGroupings[j + 1], :
        ].melt(id_vars=None)["value"]
        ModifiedData[DataPointAsColumns[j]] = input_data.loc[
            HourlyGroupings[j + 1], :
        ].melt(id_vars=None)["value"]

    # Eliminate grouping including the hour with largest system laod (GW) - this group
    # will be manually included in the outputs
    if include_peak_day:
        # IP()
        GroupingwithPeakLoad = ["p" + str(int(hr_maxSysLoad / 24 / days_in_group + 1))]
        ClusteringInputDF = ModifiedDataNormalized.drop(GroupingwithPeakLoad, axis=1)
    else:
        ClusteringInputDF = ModifiedDataNormalized

    ################################## k-means clustering process
    # create Kmeans clustering model and specify the number of clusters gathered
    # number of replications =100, squared euclidean distance

    if include_peak_day:  # If peak day in cluster, generate one less cluster
        num_clusters = num_clusters - 1

    # K-means clutering with 100 trials with randomly selected starting values
    model = KMeans(
        n_clusters=num_clusters, n_init=n_init, init="k-means++", random_state=42
    )
    model.fit(ClusteringInputDF.values.transpose())

    # Store clustered data
    # Create an empty list storing weight of each cluster
    EachClusterWeight = [None] * num_clusters

    # Create an empty list storing name of each data point
    EachClusterRepPoint = [None] * num_clusters

    # creating a dataframe for storing the mapping between representative time period and the entire year
    time_series_mapping = pd.DataFrame(columns=["Period_Index", "Rep_Period_Index"])

    for k in range(num_clusters):
        # Number of points in kth cluster (i.e. label=0)
        EachClusterWeight[k] = len(model.labels_[model.labels_ == k])

        # Compute Euclidean distance of each point from centroid of cluster k
        dist = {
            ClusteringInputDF.loc[:, model.labels_ == k].columns[j]: np.linalg.norm(
                ClusteringInputDF.loc[:, model.labels_ == k].values.transpose()[j]
                - model.cluster_centers_[k]
            )
            for j in range(EachClusterWeight[k])
        }

        # Select column name closest with the smallest euclidean distance to the mean
        EachClusterRepPoint[k] = min(dist, key=lambda k: dist[k])

        # Creating a list that matches each week to a representative week
        for j in range(EachClusterWeight[k]):
            time_series_mapping = time_series_mapping.append(
                pd.DataFrame(
                    {
                        "Period_Index": int(
                            ClusteringInputDF.loc[:, model.labels_ == k].columns[j][1:]
                        ),
                        "Rep_Period_Index": k + 1,
                    },
                    index=[0],
                ),
                ignore_index=True,
            )
    if include_peak_day:
        # appending the week representing peak load
        time_series_mapping = time_series_mapping.append(
            pd.DataFrame(
                {
                    "Period_Index": int(GroupingwithPeakLoad[0][1:]),
                    "Rep_Period_Index": k + 2,
                },
                index=[0],
            ),
            ignore_index=True,
        )

    # same CSV file that will be used in GenX
    time_series_mapping = time_series_mapping.sort_values(by=["Period_Index"])
    time_series_mapping = time_series_mapping.reset_index(drop=True)

    # extract month corresponding to each time slot
    time_series_mapping["Month"] = 0
    for Period_Index in time_series_mapping["Period_Index"]:
        dayOfYear = days_in_group * Period_Index
        d = datetime.datetime.strptime("{} {}".format(dayOfYear, 2011), "%j %Y")
        time_series_mapping["Month"][Period_Index - 1] = d.month

    # Storing selected groupings in a new data frame with appropriate dimensions
    # (E.g. load in GW)
    ClusterOutputDataTemp = ModifiedData[EachClusterRepPoint]

    # Selecting rows corresponding to Load in excluded subperiods and exclude them from
    # scale factor calculation
    NRowsLoad = len(load_col_names)
    # Excluding grouping with peak hr from scale factor calculation
    if include_peak_day:
        Actualdata = ModifiedData.loc[0 : 24 * days_in_group * NRowsLoad - 1, :].drop(
            GroupingwithPeakLoad, axis=1
        )
    else:
        Actualdata = ModifiedData.loc[0 : 24 * days_in_group * NRowsLoad - 1, :]

    # Scale factor to adjust total generation in original data set to be equal to scaled
    # up total generation in sampled data set
    SampleweeksAnnualTWh = sum(
        [
            ClusterOutputDataTemp.loc[
                0 : 24 * days_in_group * NRowsLoad - 1, EachClusterRepPoint[j]
            ].sum()
            * EachClusterWeight[j]
            for j in range(num_clusters)
        ]
    )
    ScaleFactor = (
        Actualdata.loc[0 : 24 * days_in_group * NRowsLoad - 1, :].sum().sum()
        / SampleweeksAnnualTWh
    )

    # Updated load values in GW
    ClusterOutputDataTemp.loc[0 : 24 * days_in_group * NRowsLoad - 1, :] = (
        ScaleFactor
        * ClusterOutputDataTemp.loc[0 : 24 * days_in_group * NRowsLoad - 1, :]
    )

    # Add the grouping with the peak hour back into the cluster if that is excluded in
    # the clustering
    if include_peak_day:
        EachClusterRepPoint = EachClusterRepPoint + GroupingwithPeakLoad
        EachClusterWeight = EachClusterWeight + [1]
        ClusterOutputData = pd.concat(
            [ClusterOutputDataTemp, ModifiedData[GroupingwithPeakLoad]],
            axis=1,
            sort=False,
        )
    else:
        ClusterOutputData = ClusterOutputDataTemp

    # Store weights for each selected hour  Number of days *24, for each week
    ClusteredWeights = pd.DataFrame(
        EachClusterWeight * np.ones([days_in_group * 24, len(EachClusterWeight)]),
        columns=EachClusterRepPoint,
    )

    # Storing weights in final output data column
    final_output_data["GrpWeight"] = ClusteredWeights.melt(id_vars=None)["value"]

    # Regenerating data organized by time series (columns) and representative time
    # periods (hours)
    for i in range(len(new_col_names) - 1):
        final_output_data[new_col_names[i]] = ClusterOutputData.loc[
            ConcatenatedRowNames == new_col_names[i], :
        ].melt(id_vars=None)["value"]

    # Calculating error metrics and Annual profile
    FullLengthOutputs = final_output_data
    for j in range(len(EachClusterWeight)):
        # Selecting rows of the FinalOutputData dataframe to append
        df_try = final_output_data.truncate(
            before=days_in_group * 24 * j, after=days_in_group * 24 * (j + 1) - 1
        )
        #        print(EachClusterWeight[j])
        if (
            EachClusterWeight[j] > 1
        ):  # Need to duplicate entries only weight is greater than 1
            FullLengthOutputs = FullLengthOutputs.append(
                [df_try] * (EachClusterWeight[j] - 1), ignore_index=True
            )

    #  Root mean square error between the duration curves of each time series
    # Only conisder the points consider in the k-means clustering - ignoring any days
    # dropped off from original data set  due to rounding
    RMSE = {
        col: np.linalg.norm(
            np.sort(input_data.truncate(after=len(FullLengthOutputs) - 1)[col].values)
            - np.sort(FullLengthOutputs[col].values)
        )
        for col in original_col_names
    }

    load_df = final_output_data.loc[:, load_col_names]
    # if variable_only:
    resource_df = pd.DataFrame(
        columns=resource_col_names, index=final_output_data.index
    )
    resource_df.loc[:, var_col_names] = final_output_data.loc[:, var_col_names]
    resource_df = resource_df.fillna(value=1)

    # load_df["Sub_Weights"] = np.nan
    # load_df.loc[: len(EachClusterWeight) - 1, "Sub_Weights"] = (
    #     np.array(EachClusterWeight) * NumGrpDays * 24
    # )
    # load_df.to_csv("load_time_reduced.csv", index=False)
    # renewable_df = FinalOutputData.loc[
    #     :, [col for col in FinalOutputData.columns if "Load_" not in col]
    # ]
    # renewable_df = renewable_df.drop(columns=["GrpWeight"])
    # renewable_df.insert(loc=0, column="Resource", value=renewable_df.index + 1)
    # renewable_df.to_csv("renewables_time_reduced.csv", index=False)
    rep_period_map = {i + 1: int(p[1:]) for i, p in enumerate(EachClusterRepPoint)}
    time_series_mapping["Rep_Period"] = time_series_mapping["Rep_Period_Index"].map(
        rep_period_map
    )
    EachClusterRepPoint = pd.DataFrame(EachClusterRepPoint, columns=["slot"])
    return (
        {
            "load_profiles": load_df,  # Scaled Output Load and Renewables profiles for the sampled representative groupings
            "resource_profiles": resource_df,
            "ClusterWeights": EachClusterWeight,  # Weight of each for the representative groupings
            "AnnualGenScaleFactor": ScaleFactor,  # Scale factor used to adjust load output to match annual generation of original data
            "RMSE": RMSE,  # Root mean square error between full year data and modeled full year data (duration curves)
            "AnnualProfile": FullLengthOutputs,
            "time_series_mapping": time_series_mapping,
        },
        EachClusterRepPoint,
        EachClusterWeight,
    )  # Modeled duration curves GW



import copy
import json
import os
import re
from pathlib import Path
from typing import Any, Dict, Iterable, List, Mapping, Optional, Union

import numpy as np
import pandas as pd
import pyarrow
import pyarrow.parquet as pq
import scipy.cluster.hierarchy

CAPACITY = "mw"
MERGE = {
    "sums": [CAPACITY, "area"],
    "means": [
        "lcoe",
        "interconnect_annuity",
        "interconnect_capex_mw",
        "offshore_spur_miles",
        "spur_miles",
        "tx_miles",
        "site_substation_spur_miles",
        "substation_metro_tx_miles",
        "site_metro_spur_miles",
        "m_popden",
    ],
    "weight": CAPACITY,
    "uniques": ["ipm_region", "metro_id"],
}
NREL_ATB_TECHNOLOGY_MAP = {
    ("utilitypv", None): {"technology": "utilitypv"},
    ("landbasedwind", None): {"technology": "landbasedwind"},
    ("offshorewind", None): {"technology": "offshorewind"},
    ("hydropower", None): {"technology": "hydro"},
    **{
        ("offshorewind", f"otrg{x}"): {
            "technology": "offshorewind",
            "turbine_type": "fixed",
        }
        for x in range(1, 8)
    },
    **{
        ("offshorewind", f"class{x}"): {
            "technology": "offshorewind",
            "turbine_type": "fixed",
        }
        for x in range(1, 8)
    },
    **{
        ("offshorewind", f"otrg{x}"): {
            "technology": "offshorewind",
            "turbine_type": "floating",
        }
        for x in range(8, 16)
    },
    **{
        ("offshorewind", f"class{x}"): {
            "technology": "offshorewind",
            "turbine_type": "floating",
        }
        for x in range(8, 16)
    },
}
EIA_TECHNOLOGY_MAP = {
    "conventionalhydroelectric": {"technology": "hydro", "small": False},
    "smallhydroelectric": {"technology": "hydro", "small": True},
    "onshorewindturbine": {"technology": "landbasedwind"},
    "offshorewindturbine": {"technology": "offshorewind"},
    "solarphotovoltaic": {"technology": "utilitypv"},
}


def _normalize(x: Optional[str]) -> Optional[str]:
    """
    Normalize string to lowercase, no whitespace, and no underscores.

    Examples
    --------
    >>> _normalize('Offshore Wind')
    'offshorewind'
    >>> _normalize('OffshoreWind')
    'offshorewind'
    >>> _normalize('Offshore_Wind')
    'offshorewind'
    >>> _normalize(None) is None
    True
    """
    if not x:
        return x
    return re.sub(r"\s+|_", "", x.lower())


def map_nrel_atb_technology(tech: str, detail: str = None) -> Dict[str, Any]:
    """
    Map NREL ATB technology to resource groups.

    Parameters
    ----------
    tech
        Technology.
    detail
        Technology detail.

    Returns
    -------
    dict
        Key, value pairs identifying one or more resource groups.

    Examples
    --------
    >>> map_nrel_atb_technology('UtilityPV', 'LosAngeles')
    {'technology': 'utilitypv'}
    >>> map_nrel_atb_technology('LandbasedWind', 'LTRG1')
    {'technology': 'landbasedwind'}
    >>> map_nrel_atb_technology('OffShoreWind')
    {'technology': 'offshorewind'}
    >>> map_nrel_atb_technology('OffShoreWind', 'OTRG3')
    {'technology': 'offshorewind', 'turbine_type': 'fixed'}
    >>> map_nrel_atb_technology('OffShoreWind', 'OTRG7')
    {'technology': 'offshorewind', 'turbine_type': 'floating'}
    >>> map_nrel_atb_technology('Hydropower')
    {'technology': 'hydro'}
    >>> map_nrel_atb_technology('Hydropower', 'NSD4')
    {'technology': 'hydro'}
    >>> map_nrel_atb_technology('Unknown')
    {}
    """
    tech = _normalize(tech)
    detail = _normalize(detail)
    group = {}
    for k, v in NREL_ATB_TECHNOLOGY_MAP.items():
        if (tech == k[0] or not k[0]) and (detail == k[1] or not k[1]):
            group.update(v)
    return group


def map_eia_technology(tech: str) -> Dict[str, Any]:
    """
    Map EIA technology to resource groups.

    Parameters
    ----------
    tech
        Technology.

    Returns
    -------
    dict
        Key, value pairs identifying one or more resource groups.

    Examples
    --------
    >>> map_eia_technology('Solar Photovoltaic')
    {'technology': 'utilitypv'}
    >>> map_eia_technology('solar_photovoltaic')
    {'technology': 'utilitypv'}
    >>> map_eia_technology('Onshore Wind Turbine')
    {'technology': 'landbasedwind'}
    >>> map_eia_technology('Offshore Wind Turbine')
    {'technology': 'offshorewind'}
    >>> map_eia_technology('Conventional Hydroelectric')
    {'technology': 'hydro', 'small': False}
    >>> map_eia_technology('Small Hydroelectric')
    {'technology': 'hydro', 'small': True}
    >>> map_eia_technology('Unknown')
    {}
    """
    tech = _normalize(tech)
    group = {}
    for k, v in EIA_TECHNOLOGY_MAP.items():
        if tech == k or not k:
            group.update(v)
    return group


class Table:
    """
    Cached interface for tabular data.

    Supports parquet and csv formats.

    Parameters
    ----------
    path
        Path to dataset.
    df
        In-memory dataframe.

    Attributes
    ----------
    path : Union[str, os.PathLike]
        Path to the dataset.
    df : pd.DataFrame
        Cached dataframe.
    format : str
        Dataset format ('parquet' or 'csv'), or `None` if in-memory only.
    columns : list
        Dataset column names.

    Raises
    ------
    ValueError
        Missing either path or dataframe.
    ValueError
        Dataframe columns are not all strings.

    Examples
    --------
    In-memory dataframe:

    >>> df = pd.DataFrame({'id': [1, 2], 'x': [10, 20]})
    >>> table = Table(df = df)
    >>> table.format is None
    True
    >>> table.columns
    ['id', 'x']
    >>> table.read()
       id   x
    0   1  10
    1   2  20
    >>> table.read(columns=['id'])
       id
    0   1
    1   2
    >>> table.clear()
    >>> table.df is not None
    True

    File dataset (csv):

    >>> import tempfile
    >>> fp = tempfile.NamedTemporaryFile()
    >>> df.to_csv(fp.name, index=False)
    >>> table = Table(path = fp.name)
    >>> table.format
    'csv'
    >>> table.columns
    ['id', 'x']
    >>> table.read(cache=False)
       id   x
    0   1  10
    1   2  20
    >>> table.df is None
    True
    >>> table.read(columns=['id'], cache=True)
       id
    0   1
    1   2
    >>> table.df is not None
    True
    >>> table.clear()
    >>> table.df is None
    True
    >>> fp.close()
    """

    def __init__(
        self, path: Union[str, os.PathLike] = None, df: pd.DataFrame = None
    ) -> None:
        self.path = path
        self.df = df
        if df is not None:
            if any(not isinstance(x, str) for x in df.columns):
                raise ValueError("Dataframe columns are not all strings")
        self.format = None
        self._dataset = None
        self._columns = None
        if path is not None:
            try:
                self._dataset = pq.ParquetDataset(path)
                self._columns = pq.read_schema(path).names
                self.format = "parquet"
            except pyarrow.lib.ArrowInvalid:
                # Assume CSV file
                self.format = "csv"
        if path is None and df is None:
            raise ValueError("Mising either path to tabular data or a pandas DataFrame")

    @property
    def columns(self) -> list:
        if self.df is not None:
            return list(self.df.columns)
        if self._columns is None:
            if self.format == "csv":
                self._columns = pd.read_csv(self.path, nrows=0).columns
        return list(self._columns)

    def read(self, columns: Iterable = None, cache: bool = None) -> pd.DataFrame:
        """
        Read data from memory or from disk.

        Parameters
        ----------
        columns
            Names of column to read. If `None`, all columns are read.
        cache
            Whether to cache the full dataset in memory. If `None`,
            the dataset is cached if `columns` is `None`, and not otherwise.

        Returns
        -------
        pd.DataFrame
            Data as a dataframe.
        """
        if self.df is not None:
            return self.df[columns] if columns is not None else self.df
        if cache is None:
            cache = columns is None
        read_columns = None if cache else columns
        if self.format == "csv":
            df = pd.read_csv(self.path, usecols=read_columns, dtype={"metro_id": str})
        elif self.format == "parquet":
            df = self._dataset.read(columns=read_columns).to_pandas()
        if cache:
            self.df = df
        return df[columns] if columns is not None else df

    def clear(self) -> None:
        """
        Clear the dataset cache.

        Only applies if :attr:`path` is set so that the dataset can be reread from file.
        """
        if self.path is not None:
            self.df = None


class ResourceGroup:
    """
    Group of resources sharing common attributes.

    Parameters
    ----------
    group
        Group metadata.

        - `technology` : str
          Resource type ('utilitypv', 'landbasedwind', or 'offshorewind').
        - `existing` : bool
          Whether resources are new (`False`, default) or existing (`True`).
        - `tree` : str, optional
          The name of the resource metadata attribute by
          which to differentiate between multiple precomputed hierarchical trees.
          Defaults to `None` (resource group does not represent hierarchical trees).
        - `metadata` : str, optional
          Relative path to resource metadata dataset (optional if `metadata` is `None`).
        - `profiles` : str, optional
          Relative path to resource profiles dataset.
        - ... and any additional (optional) keys.

    metadata
        Resource metadata, with one resource per row.

        - `id`: int
          Resource identifier, unique within the group.
        - `ipm_region` : str
          IPM region to which the resource delivers power.
        - `mw` : float
          Maximum resource capacity in MW.
        - `lcoe` : float, optional
          Levelized cost of energy, used to guide the selection
          (from lowest to highest) and clustering (by nearest) of resources.
          If missing, selection and clustering is by largest and nearest `mw`.

        Resources representing hierarchical trees (see `group.tree`)
        require additional attributes.

        - `parent_id` : int
          Identifier of the resource formed by clustering this resource with the one
          other resource with the same `parent_id`.
          Only resources with `level` of 1 have no `parent_id`.
        - `level` : int
          Level of tree where the resource first appears, from `m`
          (the number of resources at the base of the tree), to 1.
        - `[group.tree]` : Any
          Each unique value of this grouping attribute represents a precomputed
          hierarchical tree. When clustering resources, every tree is traversed to its
          crown before the singleton resources from the trees are clustered together.

        The following resource attributes (all float) are propagaged as:

        - weighted means (weighted by `mw`):

            - `lcoe`
            - `interconnect_annuity`
            - `tx_miles`
            - `spur_miles`
            - `offshore_spur_miles`
            - `site_substation_spur_miles`
            - `substation_metro_tx_miles`
            - `site_metro_spur_miles`

        - sums:

            - `mw`
            - `area`

        - uniques:

            - `ipm_region`
            - `metro_id`

    profiles
        Variable resource capacity profiles with normalized capacity factors
        (from 0 to 1) for every hour of the year (either 8760 or 8784 for a leap year).
        Each profile must be a column whose name matches the resource `metadata.id`.
    path
        Directory relative to which the file paths `group.metadata` and `group.profiles`
        should be read.

    Attributes
    ----------
    group : Dict[str, Any]
    metadata : Table
        Cached interface to resource metadata.
    profiles : Optional[Table]
        Cached interface to resource profiles.

    Examples
    --------
    >>> group = {'technology': 'utilitypv'}
    >>> metadata = pd.DataFrame({'id': [0, 1], 'ipm_region': ['A', 'A'], 'mw': [1, 2]})
    >>> profiles = pd.DataFrame({'0': np.full(8784, 0.1), '1': np.full(8784, 0.4)})
    >>> rg = ResourceGroup(group, metadata, profiles)
    >>> rg.test_metadata()
    >>> rg.test_profiles()
    >>> rg.get_clusters(max_clusters=1)
           ipm_region  mw                                            profile
    (1, 0)          A   3  [0.3, 0.3, 0.3, 0.3, 0.3, 0.3, 0.3, 0.3, 0.3, ...
    """

    def __init__(
        self,
        group: Dict[str, Any],
        metadata: pd.DataFrame = None,
        profiles: pd.DataFrame = None,
        path: str = ".",
        profile_path: str = None,
    ) -> None:
        from powergenome.params import SETTINGS

        self.group = {"existing": False, "tree": None, **group.copy()}

        # Convert relative paths (relative to group file) to absolute paths
        # Profiles may be stored in a single location and reused across resource groups
        if self.group.get("metadata"):
            self.group["metadata"] = Path(path) / self.group["metadata"]
        if self.group.get("profiles"):
            # Check for location passed as an arg (probably from settings file),
            # then the SETTINGS (from .env file), then assume they are in the same folder
            # as the JSON file
            if profile_path and (profile_path / self.group["profiles"]).exists():
                self.group["profiles"] = Path(profile_path) / self.group["profiles"]
            elif (
                SETTINGS.get("RESOURCE_GROUP_PROFILES")
                and (
                    Path(SETTINGS["RESOURCE_GROUP_PROFILES"]) / self.group["profiles"]
                ).exists()
            ):
                self.group["profiles"] = (
                    Path(SETTINGS["RESOURCE_GROUP_PROFILES"]) / self.group["profiles"]
                )
            else:
                self.group["profiles"] = Path(path) / self.group["profiles"]
        required = ["technology"]
        if metadata is None:
            required.append("metadata")
        missing = [key for key in required if not self.group.get(key)]
        if missing:
            raise ValueError(
                f"Group metadata missing required keys {missing}: {self.group}"
            )
        self.metadata = Table(df=metadata, path=self.group.get("metadata"))
        self.profiles = None
        if profiles is not None or self.group.get("profiles"):
            self.profiles = Table(df=profiles, path=self.group.get("profiles"))

    @classmethod
    def from_json(
        cls, path: Union[str, os.PathLike], profile_path: Union[str, os.PathLike] = None
    ) -> "ResourceGroup":
        """
        Build from JSON file.

        Parameters
        ----------
        path
            Path to JSON file.
        """
        with open(path, mode="r") as fp:
            group = json.load(fp)
        return cls(group, path=os.path.dirname(path), profile_path=profile_path)

    def test_metadata(self) -> None:
        """
        Test that `:attr:metadata` is valid.

        Raises
        ------
        ValueError
            Resource metadata missing required keys.
        """
        columns = self.metadata.columns
        required = ["ipm_region", "id", "mw"]
        if self.group.get("tree"):
            required.extend(["parent_id", "level", self.group["tree"]])
        missing = [key for key in required if key not in columns]
        if missing:
            raise ValueError(f"Resource metadata missing required keys {missing}")

    def test_profiles(self) -> None:
        """
        Test that `:attr:profiles` is valid.

        Raises
        ------
        ValueError
            Resource profiles column names do not match resource identifiers.
        ValueError
            Resource profiles are not either 8760 or 8784 elements.
        """
        if self.profiles is None:
            return None
        # Cast identifiers to string to match profile columns
        ids = self.metadata.read(columns=["id"])["id"].astype(str)
        columns = self.profiles.columns
        if not set(columns) == set(ids):
            raise ValueError(
                "Resource profiles column names do not match resource identifiers"
            )
        df = self.profiles.read(columns=columns[0])
        if len(df) not in [8760, 8784]:
            raise ValueError("Resource profiles are not either 8760 or 8784 elements")

    def get_clusters(
        self,
        ipm_regions: Iterable[str] = None,
        min_capacity: float = None,
        max_clusters: int = None,
        max_lcoe: float = None,
        cap_multiplier: float = None,
        profiles: bool = True,
        utc_offset: int = 0,
        sub_region=None,
    ) -> pd.DataFrame:
        """
        Compute resource clusters.

        Parameters
        ----------
        ipm_regions
            IPM regions in which to select resources.
            If `None`, all IPM regions are selected.
        min_capacity
            Minimum total capacity (MW). Resources are selected,
            from lowest to highest levelized cost of energy (lcoe),
            or from highest to lowest capacity if lcoe not available,
            until the minimum capacity is just exceeded.
            If `None`, all resources are selected for clustering.
        max_clusters
            Maximum number of resource clusters to compute.
            If `None`, no clustering is performed; resources are returned unchanged.
        max_lcoe
            Select only the resources with a levelized cost of electricity (lcoe)
            below this maximum. Takes precedence over `min_capacity`.
        cap_multiplier
            Multiplier applied to resource capacity before selection by `min_capacity`.
        profiles
            Whether to include cluster profiles, if available, in column `profile`.

        Returns
        -------
        pd.DataFrame
            Clustered resources whose indices are tuples of the resource identifiers
            from which they were constructed.

        Raises
        ------
        ValueError
            No resources found or selected.
        """
        df = self.metadata.read().set_index("id")
        if ipm_regions is not None:
            # Filter by IPM region
            df = df[df["ipm_region"].isin(ipm_regions)]
        if sub_region is not None:
            df = df.loc[df[self.group["sub_region"]] == sub_region, :]
        if cap_multiplier is not None:
            # Apply capacity multiplier
            df[CAPACITY] *= cap_multiplier

        if df.empty:
            drop_keys = ["tree", "metadata", "site_cluster", "profiles"]
            group_info = {k: v for k, v in self.group.items() if k not in drop_keys}
            raise KeyError(
                f"""
                No resources for the group
                {group_info}
                were found in the model region containing IPM Regions
                {ipm_regions}
            """
            )
        # Sort resources by lcoe (ascending) or capacity (descending)
        by = "lcoe" if "lcoe" in df else CAPACITY
        df = df.sort_values(by, ascending=by == "lcoe")
        # Select base resources
        tree = self.group["tree"]
        if tree:
            idx = df.index
            df = df.set_index(tree)
            max_level = df.index.map(df.groupby(tree)["level"].max())
            df = df.reset_index()
            df.index = idx
            base = (df["level"] == max_level).values
            mask = base.copy()
        else:
            mask = np.ones(len(df), dtype=bool)
        if min_capacity:
            # Select resources until min_capacity reached
            temp = (df.loc[mask, CAPACITY].cumsum() < min_capacity).values
            temp[temp.argmin()] = True
            mask[mask] = temp
        if max_lcoe and "lcoe" in df:
            # Select clusters with LCOE below the cutoff
            mask[mask] = df.loc[mask, "lcoe"] <= max_lcoe
        if not mask.any():
            raise ValueError("No resources found or selected")
        if tree:
            # Only keep trees with one ore more base resources
            if isinstance(tree, list):
                df["tree"] = ""
                for t in tree:
                    df["tree"] += df[t].astype(str)
            else:
                df["tree"] = df[tree]
            selected = (
                pd.Series(mask, index=df.index)
                .groupby(df["tree"])
                .transform(lambda x: x.sum() > 0)
            )
            # Add non-base resources to selected trees
            mask |= selected & ~base
        # Apply mask
        df = df[mask]
        # Prepare merge
        merge = copy.deepcopy(MERGE)
        # Prepare profiles
        if profiles and self.profiles is not None:
            df["profile"] = list(
                np.roll(
                    self.profiles.read(columns=df.index.astype(str)).values.T,
                    utc_offset,
                )
            )
            merge["means"].append("profile")
        # Compute clusters
        if tree:
            return cluster_trees(df, by=by, tree=tree, max_rows=max_clusters, **merge)
        return cluster_rows(df, by=df[[by]], max_rows=max_clusters, **merge)


class ClusterBuilder:
    """
    Builds clusters of resources.

    Parameters
    ----------
    groups
        Groups of resources. See :class:`ResourceGroup`.

    Attributes
    ----------
    groups : Iterable[ResourceGroup]

    Examples
    --------
    Prepare the resource groups.

    >>> groups = []
    >>> group = {'technology': 'utilitypv'}
    >>> metadata = pd.DataFrame({'id': [0, 1], 'ipm_region': ['A', 'A'], 'mw': [1, 2]})
    >>> profiles = pd.DataFrame({'0': np.full(8784, 0.1), '1': np.full(8784, 0.4)})
    >>> groups.append(ResourceGroup(group, metadata, profiles))
    >>> group = {'technology': 'utilitypv', 'existing': True}
    >>> metadata = pd.DataFrame({'id': [0, 1], 'ipm_region': ['B', 'B'], 'mw': [1, 2]})
    >>> profiles = pd.DataFrame({'0': np.full(8784, 0.1), '1': np.full(8784, 0.4)})
    >>> groups.append(ResourceGroup(group, metadata, profiles))
    >>> builder = ClusterBuilder(groups)

    Compute resource clusters.

    >>> builder.get_clusters(ipm_regions=['A'], max_clusters=1,
    ...     technology='utilitypv', existing=False)
          ids ipm_region  mw  ...         profile technology  existing
    0  (1, 0)          A   3  [0.3, 0.3, 0.3, ...  utilitypv     False
    >>> builder.get_clusters(ipm_regions=['B'], min_capacity=2,
    ...     technology='utilitypv', existing=True)
        ids ipm_region  mw  ...         profile technology  existing
    0  (1,)          B   2  [0.4, 0.4, 0.4, ...  utilitypv      True

    Errors arise if search criteria is either ambiguous or results in an empty result.

    >>> builder.get_clusters(ipm_regions=['A'], technology='utilitypv')
    Traceback (most recent call last):
      ...
    ValueError: Parameters match multiple resource groups: [{...}, {...}]
    >>> builder.get_clusters(ipm_regions=['B'], technology='utilitypv', existing=False)
    Traceback (most recent call last):
      ...
    ValueError: No resources found or selected
    """

    def __init__(self, groups: Iterable[ResourceGroup]) -> None:
        self.groups = groups

    @classmethod
    def from_json(
        cls,
        paths: Iterable[Union[str, os.PathLike]],
        profile_path: Union[str, os.PathLike] = None,
    ) -> "ClusterBuilder":
        """
        Load resources from resource group JSON files.

        Parameters
        ----------
        paths
            Paths to resource group JSON files.

        Raises
        ------
        ValueError
            No resource groups specified.
        """
        paths = list(paths)
        if not paths:
            raise ValueError(f"No resource groups specified {paths}")
        return cls([ResourceGroup.from_json(path, profile_path) for path in paths])

    def find_groups(self, **kwargs: Any) -> List[ResourceGroup]:
        """
        Return the resource groups matching the specified arguments.

        Parameters
        ----------
        **kwargs
            Parameters to match against resource group metadata.
        """
        return [
            rg
            for rg in self.groups
            if all(
                k in rg.group and rg.group[k] == v
                for k, v in kwargs.items()
                if k not in ["sub_region", "extra_uniques"]
            )
        ]

    def get_clusters(
        self,
        ipm_regions: Iterable[str] = None,
        min_capacity: float = None,
        max_clusters: int = None,
        max_lcoe: float = None,
        cap_multiplier: float = None,
        utc_offset: int = 0,
        **kwargs: Any,
    ) -> pd.DataFrame:
        """
        Compute resource clusters.

        See :meth:`ResourceGroup.get_clusters` for parameter descriptions.

        The following fields are added:

        - `ids` (tuple): Original resource identifiers.
        - **kwargs: Parameters used to uniquely identify the group.

        Parameters
        ----------
        ipm_regions
        min_capacity
        max_clusters
        max_lcoe
        cap_multiplier
        **kwargs
            Parameters to :meth:`find_groups` for selecting the resource group.

        Raises
        ------
        ValueError
            Parameters do not match any resource groups.
        ValueError
            Parameters match multiple resource groups.
        """
        groups = self.find_groups(**kwargs)
        if not groups:
            raise ValueError(f"Parameters do not match any resource groups: {kwargs}")
        if len(groups) > 1:
            meta = [rg.group for rg in groups]
            raise ValueError(f"Parameters match multiple resource groups: {meta}")
        return (
            groups[0]
            .get_clusters(
                ipm_regions=ipm_regions,
                min_capacity=min_capacity,
                max_clusters=max_clusters,
                max_lcoe=max_lcoe,
                cap_multiplier=cap_multiplier,
                utc_offset=utc_offset,
                sub_region=kwargs.get("sub_region"),
            )
            .assign(**kwargs)
            .rename_axis("ids")
            .reset_index()
        )


def _tuple(x: Any) -> tuple:
    """
    Cast object to tuple.

    Examples
    --------
    >>> _tuple(1)
    (1,)
    >>> _tuple([1])
    (1,)
    >>> _tuple('string')
    ('string',)
    """
    if np.iterable(x) and not isinstance(x, str):
        return tuple(x)
    return (x,)


def merge_row_pair(
    a: Mapping,
    b: Mapping,
    sums: Iterable = None,
    means: Iterable = None,
    weight: Any = None,
    uniques: Iterable = None,
) -> dict:
    """
    Merge two mappings into one.

    Parameters
    ----------
    a
        First mapping (e.g. :class:`dict`, :class:`pd.Series`).
    b
        Second mapping.
    means
        Keys of values to average.
    weight
        Key of values to use as weights for weighted averages.
        If `None`, averages are not weighted.
    uniques
        Keys of values for which to return the value if equal, and `None` if not.

    Returns
    -------
    dict
        Merged row as a dictionary.

    Examples
    --------
    >>> df = pd.DataFrame({'mw': [1, 2], 'area': [10, 20], 'lcoe': [0.1, 0.4]})
    >>> a, b = df.to_dict('records')
    >>> merge_row_pair(a, b, sums=['area', 'mw'], means=['lcoe'], weight='mw')
    {'area': 30, 'mw': 3, 'lcoe': 0.3}
    >>> merge_row_pair(a, b, sums=['area', 'mw'], means=['lcoe'])
    {'area': 30, 'mw': 3, 'lcoe': 0.25}
    >>> b['mw'] = 1
    >>> merge_row_pair(a, b, uniques=['mw', 'area'])
    {'mw': 1, 'area': None}
    """
    merge = {}
    if sums:
        for key in sums:
            merge[key] = a[key] + b[key]
    if means:
        if weight:
            total = a[weight] + b[weight]
            aw = a[weight] / total
            bw = b[weight] / total
        else:
            aw = 0.5
            bw = 0.5
        for key in means:
            merge[key] = a[key] * aw + b[key] * bw
    if uniques:
        for key in uniques:
            merge[key] = a[key] if a[key] == b[key] else None
    return merge


def cluster_rows(
    df: pd.DataFrame, by: Iterable[Iterable], max_rows: int = None, **kwargs: Any
) -> pd.DataFrame:
    """
    Merge rows in dataframe by hierarchical clustering.

    Uses the Ward variance minimization algorithm to incrementally merge rows.
    See :func:`scipy.cluster.hierarchy.linkage`.

    Parameters
    ----------
    df
        Rows to merge (m, ...).
    by
        2-dimensional array of observation vectors (m, ...) from which to compute
        distances between each row pair.
    max_rows
        Number of rows at which to stop merging rows.
        If `None`, no clustering is performed.
    **kwargs
        Optional parameters to :func:`merge_row_pair`.

    Returns
    -------
    pd.DataFrame
        Merged rows as a dataframe.
        Their indices are tuples of the original row indices from which they were built.
        If original indices were already iterables, they are merged
        (e.g. (1, 2) and (3, ) becomes (1, 2, 3)).

    Raises
    ------
    ValueError
        Max number of rows must be greater than zero.

    Examples
    --------
    With the default (range) row index:

    >>> df = pd.DataFrame({'mw': [1, 2, 3], 'area': [4, 5, 6], 'lcoe': [0.1, 0.4, 0.2]})
    >>> kwargs = {'sums': ['mw', 'area'], 'means': ['lcoe'], 'weight': 'mw'}
    >>> cluster_rows(df, by=df[['lcoe']], **kwargs)
          mw  area  lcoe
    (0,)   1     4   0.1
    (1,)   2     5   0.4
    (2,)   3     6   0.2
    >>> cluster_rows(df, by=df[['lcoe']], max_rows=2, **kwargs)
            mw  area   lcoe
    (1,)     2     5  0.400
    (0, 2)   4    10  0.175

    With a custom row index:

    >>> df.index = ['a', 'b', 'c']
    >>> cluster_rows(df, by=df[['lcoe']], max_rows=2, **kwargs)
            mw  area   lcoe
    (b,)     2     5  0.400
    (a, c)   4    10  0.175

    With an iterable row index:

    >>> df.index = [(1, 2), (4, ), (3, )]
    >>> cluster_rows(df, by=df[['lcoe']], max_rows=2, **kwargs)
               mw  area   lcoe
    (4,)        2     5  0.400
    (1, 2, 3)   4    10  0.175
    """
    nrows = len(df)
    if max_rows is None:
        max_rows = len(df)
    elif max_rows < 1:
        raise ValueError("Max number of rows must be greater than zero")
    drows = nrows - max_rows
    index = [_tuple(x) for x in df.index] + [None] * drows
    merge = prepare_merge(kwargs, df)
    df = df[get_merge_columns(merge, df)].reset_index(drop=True)
    if drows < 1:
        df.index = index
        return df
    # Convert dataframe rows to dictionaries
    rows = df.to_dict("records")
    # Preallocate new rows
    rows += [None] * drows
    # Preallocate new rows
    Z = scipy.cluster.hierarchy.ward(by)
    n = nrows + drows
    mask = np.ones(n, dtype=bool)
    for i, link in enumerate(Z[:drows, 0:2].astype(int)):
        mask[link] = False
        pid = nrows + i
        rows[pid] = merge_row_pair(rows[link[0]], rows[link[1]], **merge)
        index[pid] = index[link[0]] + index[link[1]]
    clusters = pd.DataFrame([x for x, m in zip(rows, mask) if m])
    # Preserve original column order
    clusters = clusters[[x for x in df.columns if x in clusters]]
    clusters.index = [x for x, m in zip(index, mask) if m]
    return clusters


def build_tree(
    df: pd.DataFrame, by: Iterable[Iterable], max_level: int = None, **kwargs: Any
) -> pd.DataFrame:
    """
    Build a hierarchical tree of rows in a dataframe.

    Uses the Ward variance minimization algorithm to incrementally merge rows.
    See :func:`scipy.cluster.hierarchy.linkage`.

    Parameters
    ----------
    df
        Rows to merge (m, ...).
        Should not have columns `id`, `parent_id`, and `level`, as these are appended to
        the result dataframe.
    by
        2-dimensional array of observation vectors (m, ...) from which to compute
        distances between each row pair.
    max_level
        Maximum level of tree to return,
        from m (the number of rows in `df`, if `None`) to 1.
    **kwargs
        Optional parameters to :func:`merge_row_pair`.

    Returns
    -------
    pd.DataFrame
        Hierarchical tree as a dataframe.
        Row indices are tuples of the original row indices from which they were built.
        If original indices were already iterables, they are merged
        (e.g. (1, 2) and (3, ) becomes (1, 2, 3)).
        The following columns are added:

        - `id` (int): New row identifier (0, ..., 0 + n).
        - `parent_id` (Int64): New row identifer of parent row.
        - `level` (int): Tree level of row (max_level, ..., 1).

    Raises
    ------
    ValueError
        Max level of tree must be greater than zero.

    Examples
    --------
    >>> df = pd.DataFrame({'mw': [1, 2, 3], 'area': [4, 5, 6], 'lcoe': [0.1, 0.4, 0.2]})
    >>> kwargs = {'sums': ['area', 'mw'], 'means': ['lcoe'], 'weight': 'mw'}
    >>> build_tree(df, by=df[['lcoe']], **kwargs)
               mw  area   lcoe  id  parent_id  level
    (0,)        1     4  0.100   0          3      3
    (1,)        2     5  0.400   1          4      3
    (2,)        3     6  0.200   2          3      3
    (0, 2)      4    10  0.175   3          4      2
    (1, 0, 2)   6    15  0.250   4        NaN      1
    >>> build_tree(df, by=df[['lcoe']], max_level=2, **kwargs)
               mw  area   lcoe  id  parent_id  level
    (1,)        2     5  0.400   0          2      2
    (0, 2)      4    10  0.175   1          2      2
    (1, 0, 2)   6    15  0.250   2        NaN      1
    >>> build_tree(df, by=df[['lcoe']], max_level=1, **kwargs)
               mw  area  lcoe  id  parent_id  level
    (1, 0, 2)   6    15  0.25   0        NaN      1
    """
    nrows = len(df)
    if max_level is None:
        max_level = nrows
    else:
        max_level = min(max_level, nrows)
        if max_level < 1:
            raise ValueError("Max level of tree must be greater than zero")
    drows = nrows - 1
    index = [_tuple(x) for x in df.index] + [None] * drows
    df = df.reset_index(drop=True)
    merge = prepare_merge(kwargs, df)
    columns = get_merge_columns(merge, df)
    df = df[columns]
    if drows < 1:
        df.index = index
        return df
    # Convert dataframe rows to dictionaries
    rows = df.to_dict("records")
    # Preallocate new rows
    rows += [None] * drows
    Z = scipy.cluster.hierarchy.linkage(by, method="ward")
    n = nrows + drows
    mask = np.ones(n, dtype=bool)
    level = np.concatenate((np.full(nrows, nrows), np.arange(drows, 0, -1)))
    parent_id = np.zeros(n)
    drop = nrows - max_level
    for i, link in enumerate(Z[:, 0:2].astype(int)):
        if i < drop:
            mask[link] = False
        pid = nrows + i
        parent_id[link] = pid
        rows[pid] = merge_row_pair(rows[link[0]], rows[link[1]], **merge)
        index[pid] = index[link[0]] + index[link[1]]
    tree = pd.DataFrame([x for x, m in zip(rows, mask) if m])
    # Restore original column order
    tree = tree[columns]
    # Normalize ids to 0, ..., n
    old_ids = np.where(mask)[0]
    new_ids = np.arange(len(old_ids))
    new_parent_ids = pd.Series(np.searchsorted(old_ids, parent_id[mask]), dtype="Int64")
    new_parent_ids.iloc[-1] = np.nan
    # Bump lower levels to max_level
    level = level[mask]
    if max_level < nrows:
        stop = level.size - np.searchsorted(level[::-1], max_level, side="right")
        level[:stop] = max_level
    tree = tree.assign(id=new_ids, parent_id=new_parent_ids, level=level)
    tree.index = [x for x, m in zip(index, mask) if m]
    return tree


def cluster_trees(
    df: pd.DataFrame, by: str, tree: str = None, max_rows: int = None, **kwargs: Any
) -> pd.DataFrame:
    """
    Merge rows in a dataframe following precomputed hierarchical trees.

    Parameters
    ----------
    df
        Rows to merge.
        Must have columns `parent_id` (matching values in index), `level`, and
        the columns named in **by** and **tree**.
    by
        Name of column to use for determining merge order.
        Children with the smallest pairwise distance on this column are merged first.
    tree
        Name of column to use for differentiating between hierarchical trees.
        If `None`, assumes rows represent a single tree.
    max_rows
        Number of rows at which to stop merging rows.
        If smaller than the number of trees, :func:`cluster_rows` is used to merge
        tree heads.
        If `None`, no merging is performed and only the base rows are returned.
    **kwargs
        Optional parameters to :func:`merge_row_pair`.

    Returns
    -------
    pd.DataFrame
        Merged rows as a dataframe.
        Their indices are tuples of the original row indices from which they were built.
        If original indices were already iterables, they are merged
        (e.g. (1, 2) and (3, ) becomes (1, 2, 3)).

    Raises
    ------
    ValueError
        Max number of rows must be greater than zero.
    ValueError
        Missing required fields.
    ValueError
        `by` column not included in row merge arguments (`kwargs`).

    Examples
    --------
    >>> df = pd.DataFrame({
    ...     'level': [3, 3, 3, 2, 1],
    ...     'parent_id': pd.Series([3, 3, 4, 4, float('nan')], dtype='Int64'),
    ...     'mw': [0.1, 0.1, 0.1, 0.2, 0.3],
    ...     'area': [1, 1, 1, 2, 3]
    ... }, index=[0, 1, 2, 3, 4])
    >>> cluster_trees(df, by='mw', sums=['mw', 'area'], max_rows=2)
             mw  area
    (2,)    0.1     1
    (0, 1)  0.2     2
    >>> cluster_trees(df, by='mw', sums=['mw'], max_rows=1)
                mw
    (2, 0, 1)  0.3
    >>> cluster_trees(df, by='mw', sums=['mw'])
           mw
    (0,)  0.1
    (1,)  0.1
    (2,)  0.1
    """
    required = ["parent_id", "level", by]
    if tree:
        if isinstance(tree, list):
            required.extend(tree)
        else:
            required.append(tree)
    missing = [key for key in required if key not in df]
    if missing:
        raise ValueError(f"Missing required fields {missing}")
    if tree:
        idx = df.index
        df = df.set_index(tree)
        mask = df["level"] == df.index.map(df.groupby(tree)["level"].max())
        mask.index = idx
        df = df.reset_index()
        df.index = idx
    else:
        mask = df["level"] == df["level"].max()
    nrows = mask.sum()
    if max_rows is None:
        max_rows = nrows
    elif max_rows < 1:
        raise ValueError("Max number of rows must be greater than zero")
    merge = prepare_merge(kwargs, df)
    columns = get_merge_columns(merge, df)
    if by not in columns:
        raise ValueError(f"{by} not included in row merge arguments")
    drows = nrows - max_rows
    if drows < 1:
        df = df.loc[mask, columns].copy()
        df.index = [_tuple(x) for x in df.index]
        return df
    df = df[set(columns + required)].assign(
        _id=df.index, _ids=[_tuple(x) for x in df.index], _mask=mask
    )
    diff = lambda x: abs(x.max() - x.min())
    while drows > 0:
        # Sort parents by ascending distance of children
        # NOTE: Inefficient to recompute for all parents every time
        parents = (
            df[df["_mask"]]
            .groupby("parent_id", sort=False)
            .agg(ids=("_id", list), n=("_id", "count"), distance=(by, diff))
            .sort_values(["n", "distance"], ascending=[False, True])
        )
        if parents.empty:
            break
        if parents["n"].iloc[0] == 2:
            # Choose complete parent with lowest distance of children
            pid = parents.index[0]
            ids = parents["ids"].iloc[0]
            children = df.loc[ids].to_dict("records")
            # Compute parent
            parent = {
                # Initial attributes
                # Can access series because all columns integer
                **df.loc[pid, ["_id", "parent_id", "level"]],
                # Merged children attributes
                # NOTE: Needed only if a child is incomplete
                **merge_row_pair(children[0], children[1], **merge),
                # Indices of all past children
                "_ids": df.loc[ids[0], "_ids"] + df.loc[ids[1], "_ids"],
                "_mask": True,
            }
            # Add parent
            df.loc[pid] = pd.Series(parent, dtype=object)
            # Drop children
            df.loc[ids, "_mask"] = False
            # Decrement rows
            drows -= 1
        else:
            # Promote child with deepest parent
            parent_id = df.loc[parents.index, "level"].idxmax()
            child_id = parents.loc[parent_id, "ids"][0]
            # Update child
            tree_columns = ["_id", "parent_id", "level"]
            df.loc[child_id, tree_columns] = df.loc[parent_id, tree_columns]
            # Update index
            df.rename(index={child_id: parent_id, parent_id: np.nan}, inplace=True)
    # Apply mask
    df = df[df["_mask"]]
    # Drop temporary columns
    df.index = df["_ids"].values
    df = df.drop(columns=["_id", "_ids", "_mask"])
    if len(df) > max_rows:
        df = cluster_rows(df, by=df[[by]], max_rows=max_rows, **kwargs)
    return df[columns]


def group_rows(
    df: pd.DataFrame, ids: Iterable[Iterable]
) -> pd.core.groupby.DataFrameGroupBy:
    """
    Group dataframe rows by index.

    Parameters
    ----------
    df
        Dataframe to group.
    ids
        Groups of rows indices.

    Returns
    -------
    pd.core.groupby.DataFrameGroupBy
        Rows of `df` grouped by their membership in each index group.

    Examples
    --------
    >>> df = pd.DataFrame({'x': [2, 1, 3]}, index=[2, 1, 3])
    >>> group_rows(df, [(1, ), (2, 3), (1, 2, 3)]).sum()
       x
    0  1
    1  5
    2  6
    """
    groups = np.repeat(np.arange(len(ids)), [len(x) for x in ids])
    index = np.concatenate(ids)
    return df.loc[index].groupby(groups, sort=False)


def prune_tree(df: pd.DataFrame, level: int) -> pd.DataFrame:
    """
    Prune base levels of hierarchical tree.

    Parameters
    ----------
    df
        Dataframe representing a hierarchical tree.
        Must have columns `id`, `parent_id` and `level`.
    level
        Level at which to prune tree.

    Returns
    -------
    pd.DataFrame
        Pruned hierarchical tree.
        Column `id` (and `parent_id`) is reset to (0, ..., nrows - 1).

    Examples
    --------
    >>> parent_id = pd.Series([3, 3, 4, 4, None], dtype='Int64')
    >>> df = pd.DataFrame({
    ...     'id': [0, 1, 2, 3, 4],
    ...     'parent_id': parent_id,
    ...     'level': [3, 3, 3, 2, 1]
    ... })
    >>> prune_tree(df, level=2)
       id  parent_id  level
    2   0          2      2
    3   1          2      2
    4   2        NaN      1
    """
    levels = df["level"].max()
    if level > levels:
        return df
    # Drop direct children of all parents up to and including max_level
    pids = df["id"][(df["level"] >= level) & (df["level"] < levels)]
    df = df[~df["parent_id"].isin(pids)].copy()
    # Bump level of remaining children
    df.loc[df["level"] > level, "level"] = level
    # Normalize ids to 0, ..., n
    mask = ~df["parent_id"].isna()
    df.loc[mask, "parent_id"] = np.searchsorted(df["id"], df["parent_id"][mask])
    df["id"] = np.arange(len(df))
    return df


def prepare_merge(merge: dict, df: pd.DataFrame) -> dict:
    """
    Prepare merge for a target dataframe.

    Parameters
    ----------
    merge
        Parameters to :func:`merge_row_pair`.
    df
        Dataframe to prepare merge for.

    Raises
    ------
    ValueError
        Column names duplicated in merge.
    ValueError
        Weights not present in dataframe.
    ValueError
        Weights not included in merge.

    Examples
    --------
    >>> df = pd.DataFrame(columns=['mw', 'lcoe'])
    >>> merge = {'sums': ['mw', 'area'], 'means': ['lcoe'], 'weight': 'mw'}
    >>> prepare_merge(merge, df)
    {'sums': ['mw'], 'means': ['lcoe'], 'weight': 'mw'}
    """
    reduced = {}
    for key in "sums", "means", "uniques":
        if merge.get(key):
            reduced[key] = [x for x in merge[key] if x in df]
    columns = get_merge_columns(reduced)
    if reduced.get("means") and merge.get("weight"):
        weight = merge["weight"]
        if weight not in df:
            raise ValueError(f"Weights {weight} not present in dataframe")
        if weight not in columns:
            raise ValueError(f"Weights {weight} not included in merge")
        reduced["weight"] = weight
    return reduced


def get_merge_columns(merge: dict, df: pd.DataFrame = None) -> list:
    """
    Get columns included in merge.

    Parameters
    ----------
    merge
        Parameters to :func:`merge_row_pair`.
    df
        Dataframe.
        If provided, only matching column names are returned, in order of appearance.

    Raises
    ------
    ValueError
        Column names duplicated in merge.

    Examples
    --------
    >>> merge = {'sums': ['mw'], 'means': ['lcoe'], 'uniques': None, 'weight': 'lcoe'}
    >>> get_merge_columns(merge)
    ['mw', 'lcoe']
    >>> get_merge_columns(merge, pd.DataFrame(columns=['lcoe', 'mw']))
    ['lcoe', 'mw']
    >>> get_merge_columns({'sums': ['mw'], 'means': ['mw']})
    Traceback (most recent call last):
      ...
    ValueError: Column names duplicated in merge
    """
    columns = (
        (merge.get("sums") or [])
        + (merge.get("means") or [])
        + (merge.get("uniques") or [])
    )
    if len(columns) > len(set(columns)):
        visited = set()
        dup = [x for x in columns if x in visited or (visited.add(x) or False)]
        raise ValueError(f"Column names {dup} duplicated in merge")
    if df is not None:
        return [x for x in df if x in columns]
    return columns



"""
Transmission constraints between regions and line distance
"""

import itertools
import logging
from math import asin, cos, radians, sin, sqrt
from pathlib import Path

import pandas as pd
import sqlalchemy as sa

from powergenome.util import find_centroid, map_agg_region_names, reverse_dict_of_lists

logger = logging.getLogger(__name__)


def agg_transmission_constraints(
    pg_engine: sa.engine.base.Engine,
    settings: dict,
    pg_table: str = "transmission_single_epaipm",
    settings_agg_key: str = "region_aggregations",
) -> pd.DataFrame:
    """Aggregate transmission constraints/capacity between model regions

    Model regions can consist of one or more individual regions. When two or more regions
    are in a model region, the transmission capacity between individual regions is
    combined.

    Values in a user transmission table will override the database transmission table.

    Parameters
    ----------
    pg_engine : sa.engine.base.Engine
        Engine to conect with a database
    settings : dict
        Dictionary of settings parameters. Must include "model_regions". Optional parameters
        include "tx_value_col" (name of the data column in `pg_table`, default value is
        "firm_ttc_mw"), "user_transmission_constraints_fn" (name of user data file, must
        be combined with the parameter "input_folder"), and the value of `settings_agg_key`
        if any regions are aggregated.
    pg_table : str, optional
        Name of the database table with transmission capacity, by default "transmission_single_epaipm"
    settings_agg_key : str, optional
        Name of the settings parameter where regions are aggregated, by default "region_aggregations"

    Returns
    -------
    pd.DataFrame
        Network lines connecting regions and the min/max flow of each line (if different
        based on direction)

    Raises
    ------
    KeyError
        The specified data column is not in the database transmission table
    KeyError
        The database transmission table has duplicate lines in the same direction
    KeyError
        The specified data column is not in the user supplied transmission table
    KeyError
        The user transmission table has duplicate lines in the same direction
    """
    tx_value_col = settings.get("tx_value_col")
    if not tx_value_col:
        logger.warning(
            "No transmission value column (e.g. firm vs non-firm) was specified in the "
            "settings. The column 'firm_ttc_mw' will be used as a default. This is a change "
            "from previous versions of PG, where 'nonfirm_ttc_mw' was used. Firm transmission "
            "capacity is lower or equal to non-firm capacity."
        )
        tx_value_col = "firm_ttc_mw"
    zones = settings["model_regions"]
    zone_num_map = {
        zone: f"z{number + 1}" for zone, number in zip(zones, range(len(zones)))
    }

    combos = list(itertools.combinations(zones, 2))
    reverse_combos = [(combo[-1], combo[0]) for combo in combos]

    logger.debug("Loading transmission constraints from the database")
    transmission_constraints_table = pd.read_sql_table(pg_table, con=pg_engine)

    if tx_value_col not in transmission_constraints_table.columns:
        raise KeyError(
            f"There is no column {tx_value_col} in the transmission capacity table '{pg_table}'"
        )
    if transmission_constraints_table.duplicated(
        subset=["region_from", "region_to"]
    ).any():
        dup_lines = transmission_constraints_table.loc[
            transmission_constraints_table.duplicated(
                subset=["region_from", "region_to"]
            ),
            ["region_from", "region_to"],
        ]

        raise KeyError(
            "The transmission table has duplicate lines. This table should only have unique lines.\n",
            dup_lines,
        )
    if settings.get("user_transmission_constraints_fn"):
        logger.debug("Adding user-supplied transmission constraint data")
        user_tx_constraints = pd.read_csv(
            Path(settings["input_folder"])
            / settings["user_transmission_constraints_fn"]
        )
        if tx_value_col not in user_tx_constraints.columns:
            raise KeyError(
                f"There is no column {tx_value_col} in the user supplied transmission capacity table"
            )
        if user_tx_constraints.duplicated(subset=["region_from", "region_to"]).any():
            dup_lines = user_tx_constraints.loc[
                user_tx_constraints.duplicated(subset=["region_from", "region_to"]),
                ["region_from", "region_to"],
            ]

            raise KeyError(
                "The user transmission table has duplicate lines. This table should only have unique lines.\n",
                dup_lines,
            )

        # user constraints are needed bidirectionaly
        transmission_constraints_table = pd.concat(
            [
                transmission_constraints_table,
                user_tx_constraints,
                user_tx_constraints.rename(
                    columns={"region_from": "region_to", "region_to": "region_from"}
                ),
            ]
        )

        if transmission_constraints_table.duplicated(
            subset=["region_from", "region_to"]
        ).any():
            logger.warning(
                "The user transmission capacity table duplicates some values from the "
                "database. Database values will be discarded in these cases."
            )

        transmission_constraints_table = transmission_constraints_table.drop_duplicates(
            keep="last"
        )

    # Settings has a dictionary of lists for regional aggregations. Need
    # to reverse this to use in a map method.
    region_agg_map = reverse_dict_of_lists(settings.get(settings_agg_key))

    # IPM regions to keep. Regions not in this list will be dropped from the
    # dataframe
    keep_regions = [
        x
        for x in settings["model_regions"] + list(region_agg_map)
        if x not in region_agg_map.values()
    ]

    # Create new column "model_region_from"  and "model_region_to" with labels that
    # we're using for aggregated regions
    transmission_constraints_table = transmission_constraints_table.loc[
        (transmission_constraints_table.region_from.isin(keep_regions))
        & (transmission_constraints_table.region_to.isin(keep_regions)),
        :,
    ].drop(columns="id", errors="ignore")

    logger.debug("Map and aggregate region names for transmission constraints")
    for col in ["region_from", "region_to"]:
        model_col = "model_" + col

        transmission_constraints_table = map_agg_region_names(
            df=transmission_constraints_table,
            region_agg_map=region_agg_map,
            original_col_name=col,
            new_col_name=model_col,
        )

    keep_cols = ["model_region_from", "model_region_to", tx_value_col]
    drop_cols = [
        c for c in transmission_constraints_table.columns if c not in keep_cols
    ]
    transmission_constraints_table.drop(columns=drop_cols, inplace=True)
    transmission_constraints_table = transmission_constraints_table.groupby(
        ["model_region_from", "model_region_to"]
    ).sum()

    # Build the final output dataframe
    logger.debug(
        "Build a new transmission constraints dataframe with a single line between "
        "regions"
    )
    tc_joined = pd.DataFrame(
        columns=["Network_Lines"] + zones + ["Line_Max_Flow_MW", "Line_Min_Flow_MW"],
        index=transmission_constraints_table.reindex(combos).dropna().index,
        data=0,
    )

    if tc_joined.empty:
        logger.warning(f"No transmission lines exist between model regions {combos}")
        tc_joined["transmission_path_name"] = None
        tc_joined.rename(columns=zone_num_map, inplace=True)
        return tc_joined.reset_index(drop=True)

    tc_joined["Network_Lines"] = range(1, len(tc_joined) + 1)
    tc_joined["Line_Max_Flow_MW"] = transmission_constraints_table.reindex(
        combos
    ).dropna()

    reverse_tc = transmission_constraints_table.reindex(reverse_combos).dropna() * -1
    reverse_tc.index = tc_joined.index
    tc_joined["Line_Min_Flow_MW"] = reverse_tc

    for idx, _ in tc_joined.iterrows():
        tc_joined.loc[idx, idx[0]] = 1
        tc_joined.loc[idx, idx[-1]] = -1

    tc_joined.rename(columns=zone_num_map, inplace=True)
    tc_joined = tc_joined.reset_index()
    tc_joined["transmission_path_name"] = (
        tc_joined["model_region_from"] + "_to_" + tc_joined["model_region_to"]
    )
    # tc_joined = tc_joined.set_index("Transmission Path Name")
    tc_joined.drop(columns=["model_region_from", "model_region_to"], inplace=True)

    return tc_joined


def haversine(lon1, lat1, lon2, lat2, units="mile"):
    """
    Calculate the great circle distance between two points
    on the earth (specified in decimal degrees)

    https://gis.stackexchange.com/questions/166820/geopandas-return-lat-and-long-of-a-centroid-point
    """
    # convert decimal degrees to radians
    lon1, lat1, lon2, lat2 = map(radians, [lon1, lat1, lon2, lat2])

    # haversine formula
    dlon = lon2 - lon1
    dlat = lat2 - lat1
    a = sin(dlat / 2) ** 2 + cos(lat1) * cos(lat2) * sin(dlon / 2) ** 2
    c = 2 * asin(sqrt(a))

    if units == "mile":
        r = 3956  # Radius of earth in miles. Use 6371 for kilometers, 3956 for miles
    elif units == "km":
        r = 6371
    else:
        raise ValueError(f"Units are {units}, but should be 'mile' or 'km'")

    return c * r


def getXY(pt):
    "Return the X and Y parts of a coordinate point"
    return (pt.x, pt.y)


def single_line_distance(line_name, region_centroids, units):
    """Calculate the transmission line distance between centroids of two regions.

    Parameters
    ----------
    line_name : str
        Two region names in the format <start>_to_<end>
    region_centroids : geoseries
        Centroid points of each region with region names as the index
    units : str
        Name of the distance units to use. Options are 'mile' or 'km'.

    Returns
    -------
    float
        The distance
    """

    start, end = line_name.split("_to_")
    start_lon, start_lat = getXY(region_centroids[start])
    end_lon, end_lat = getXY(region_centroids[end])
    distance = haversine(start_lon, start_lat, end_lon, end_lat, units=units)

    return distance


def transmission_line_distance(
    trans_constraints_df, ipm_shapefile, settings, units="mile"
):
    logger.debug("Calculating transmission line distance")
    ipm_shapefile["geometry"] = ipm_shapefile.buffer(0.01)
    model_polygons = ipm_shapefile.dissolve(by="model_region")
    model_polygons = model_polygons.to_crs(epsg=4326)
    region_centroids = find_centroid(model_polygons)

    distances = [
        single_line_distance(line_name, region_centroids, units=units)
        for line_name in trans_constraints_df["transmission_path_name"]
    ]
    trans_constraints_df[f"distance_{units}"] = distances
    trans_constraints_df[f"distance_{units}"] = trans_constraints_df[
        f"distance_{units}"
    ]

    return trans_constraints_df
